#!/usr/bin/perl
#
#
# Program:	gtml
#
# Version:	1.0	 8-Jun-96	First release.
#		1.01	14-Jun-96	Added #newdefine.
#					Allow undefined constants.
#		1.02	14-Jul-96	Support path names in project file.
#		1.03	 1-Sep-96	No quotation marks in #include
#					file name is OK.
#					Allow ../ to higher levels than
#					starting directory.
#		2.0	24-Nov-96	Add lots of new features.
#		2.01	30-Nov-96	Commands can be inside comments.
#					User-specified substitute delimiters.
#               2.02    10-Feb-97	Fixed #ifdef nesting.
#					(Alan B. Chamberlin)
#		2.1	23-Jul-97	Allow subs in #include.
#					Don't downshift file names.
#					(Suggestions from Mike Sendall).
#		2.2	15-Aug-97	"allsource" command in project file.
#		2.21	 2-Feb-98	Simplify auto-generated navigation links.
#		2.22	25-Jul-98	Handles years past 2000.
#
# Description:	gtml is a program to manage groups of HTML files with
#		similar properties.
#
# Author:	Gihan Perera (gihan@pobox.com)
#		First Step Communications
#
#  Documentation and updates for this program are available at:
#
#     http://www.firststep.com.au/software
#
#  Copying and Distribution
#
#     Permission is hereby granted to copy and distribute this program
#     freely, as long as the "Author:" section above is retained. Since
#     this is provided free of charge, it comes with no warranty nor
#     guarantee of fitness for purpose.
#
#     In the spirit in which this is given, if you decide to pass this
#     on to others, either in original form or changed, you must not
#     charge any money for doing so.
#
# ---------------------------------------------------------------------

$extProj = "gtp";
$extSrc  = "gtm";
$extTgt  = "html";

$delim1 = "<<";
$delim2 = ">>";

# ---------------------------------------------------------------------

$filenum = 3;
# $suppress = 0;							#abc
$supp = 0;								#abc
@suppress = (0);							#abc
$cur = 0;								#abc
$literal = 0;
$entities = 0;

&TimeNow;

foreach $i ( @ARGV )
{
    if ( $i =~ /^-D/ )
    {
	$i =~ s/^-D//;
	( $key, $value ) = split ( /=/, $i, 2 );
	&Define ( $key, $value );
    }
    else
    {
    	if ( $i =~ /\.$extProj$/i )
    	{
	    &ProcessProjectFile ( $i );
    	}
    	elsif ( $i =~ /\.$extSrc$/i )
    	{
	    &ProcessSourceFile ( $i );
    	}
    	else
    	{
	    print "--- Skipping $i (unknown file type) ---\n";
    	}
    }
}

# ---------------------------------------------------------------------

sub ProcessProjectFile
{
    local ( $name ) = @_;

    print "--- Project file $name ---\n";

    open ( PROJECT, $name );

    while ( <PROJECT> )
    {
	chop;

#	Skip blank and comment lines.

	next if /^\/\//;
	next if /^[ \t]*$/;

#	Named constants can be defined here.

	if ( /^define[ \t]/ )
	{
	    ( $dummy, $key, $value ) = split ( /[ \t]/, $_, 3 );
	    &Define ( $key, $value );
	}
	elsif ( /^newdefine[ \t]/ )
	{
	    ( $dummy, $key, $value ) = split ( /[ \t]/, $_, 3 );
	    &Define ($key, $value) if &GetValue($key) eq "";
	}

#	Timestamp format can be defined here.

	elsif ( /^timestamp[ \t]/ )
	{
	    ( $dummy, $stamp ) = split ( /[ \t]/, $_, 2 );

	    &FormatTimestamp;
	    &Define ( "TIMESTAMP", $stamp );
	}

#	They can ask for all source files here.

        elsif ( /^allsource/ )
        {
	  foreach $file ( &AllSourceFiles() )
	  {
	    &ProcessSourceFile($file);
	  }
	}

#	Everything else is a source file name.

	else
	{
	    s/\s+/ /g;
	    ($file,$level,$title) = split(/\s/,$_,3);

	    if ( $file =~ /\.$extSrc$/i )
	    {
		if ( $level eq "" )
	    	{
		    &ProcessSourceFile ( $file );
		}
		else
		{
		    $pfile[$n_file] = $file;
		    $plevel[$n_file] = $level;
		    $ptitle[$n_file] = $title;

		    ++$n_file;
		}
	    }
	    else
	    {
		print "Skipping $_ (unknown file type)\n";
	    }
	}
    }

    for ( $ii = 0; $ii < $n_file; ++$ii )
    {
	&SetLinks();
	&ProcessSourceFile($pfile[$ii], " ($plevel[$ii])");
    }

    close PROJECT;
}

# ---------------------------------------------------------------------

sub ProcessSourceFile
{
    local($name, $level) = @_;
    local($oname);

    %saveDefines = %defines;

    print "--- $name$level\n";

    if ( ! -r $name )
    {
	print "   *** Unable to read ", $name, "\n";
    }
    else
    {
	$oname = &ResolveOutputName($name);

	die "   *** Source $name same as target $oname" if $name eq $oname;

	&GetBasename ( $oname );

	&Define ( "FILENAME", $outputName );

	open ( OUTFILE, ">$oname" );
	&ProcessLines ( $name );
	close OUTFILE;
    }
    %defines = %saveDefines;
}

# ---------------------------------------------------------------------

sub ResolveOutputName
{
    local ($f) = @_;

    $f =~ s/\.$extSrc$/\.$extTgt/;

    $f = $outputDir . "/" . $f if $outputDir ne "" && $f !~ /^\//;

#   Make sure the directory exists for the output file.

    $n = 0;

    while ( $n != -1 )
    {
	$n = index($f,"/",$n);

	$n = 1 if $n == 0;

	if ( $n != -1 )
	{
	    $dir = substr($f,0,$n);

	    mkdir($dir,"rw") if ! -d $dir;
	    ++$n;
	}
    }
    return $f;
}

# ---------------------------------------------------------------------
#
# Get the file part of the output file name and remember it.

sub GetBasename
{
    ( $outputName ) = @_;

###    $outputName =~ y/A-Z/a-z/;	# Convert to lower case
    $outputName =~ s/\\/\//g;	# "\" -> "/"

    $n = rindex ( $outputName, "/" );

    $outputName = substr ( $outputName, $n+1 ) if $n != -1;
}

# ---------------------------------------------------------------------

sub ProcessLines
{
    local ( $iname ) = @_;
    local ( $INFILE ) = $filenum++;
    local ( $saveBase ) = $basePath;
    local ( $saveTab ) = $tab;

    $iname = &ResolvePath ( $iname );

    $tab = $tab . "    ";

    if ( ! -r $iname )
    {
	print "    *** ERROR - Can't read $iname ***\n";
    }

    open ( $INFILE, $iname );

    while ( <$INFILE> )
    {
#	Allow GTML commands inside HTML comments.
	if ( /<!-- ###/ )
	{
	    s/<!-- ##//;
	    s/|-->.*$//;
	    s/\s*-->.*$//;
	}

#	Parse '#literal' command because if literal processing is ON,
#	we simply print the line and continue to the next line.

	if ( /^#literal/ )
	{
	    chop;
	    ( $dummy, $switch ) = split ( /[ \t]+/, $_, 2 );

	    if ( $switch eq "ON" )
	    {
		$literal = 1;
		next;
	    }
	    elsif ( $switch eq "OFF" )
	    {
		$literal = 0;
		next;							#abc
	    }
	    else
	    {
		&Error("Expecting #literal as 'ON' or 'OFF'");
	    }
	}

	if ( $literal )
	{
	    &Substitute();
	    print OUTFILE $_;
	    next;
	}

#	Next, parse the #ifdef/else/endif to decide if we want to suppress
#	any lines.

	if ( /^#ifdef/ || /^#ifndef/ )
	{
	    chop;
	    ( $dummy, $var ) = split ( /[ \t]+/, $_, 2 );
	    push(@iflevel,$supp);
	    $supp = ( &GetValue($var) eq "" );
	    $supp = ! $supp if /^#ifndef/;
	    push(@suppress,$supp);					#abc
            $cur = $#suppress if (! $suppress[$cur]); 			#abc
            next;							#abc
	}
	elsif ( /^#else/ )
	{
	    $suppress[$#suppress] = ! $suppress[$#suppress];
            $cur = $#suppress if (! $suppress[$cur]); 			#abc
            next;							#abc
	}
	elsif ( /^#endif/ )
	{
	    if ( $#iflevel < 0 )
	    {
		&Error("Unmatched #endif");
		$supp = 0;
                $cur = -1;						#abc
	    }
	    else
	    {
		$supp = pop(@iflevel);
		$supp = pop(@suppress);
                $cur = $#suppress if ($#suppress < $cur); 		#abc
	    }
            next;							#abc
	}

	next if $suppress[$cur]; 					#abc

#	Now do the other commands.

	if ( /^#entities/ )
	{
	    chop;
	    ( $dummy, $switch ) = split ( /[ \t]+/, $_, 2 );

	    if ( $switch eq "ON" )
	    {
		$entities = 1;
	    }
	    elsif ( $switch eq "OFF" )
	    {
		$entities = 0;
	    }
	    else
	    {
		&Error("Expecting #entities as 'ON' or 'OFF'");
	    }
	}
	elsif ( /^#include/ )
	{
	    chop;
	    &Substitute();
	    s/^#include[ \t]*"*//;
	    s/"*[ \t]*$//;

	    $f = &ResolveIncludeFile ( $_ );

	    if ( $f ne "" )
	    {
		&ProcessLines ( $f );
	    }
	}
	elsif ( /^#define/ )
	{
	    chop;
	    ( $dummy, $key, $value ) = split ( /[ \t]+/, $_, 3 );
	    &Define ( "$key", $value );
	}
	elsif ( /^#undef/ )
	{
	    chop;
	    ( $dummy, $key ) = split ( /[ \t]+/, $_, 2 );
	    &Undefine("$key");
	}
	elsif ( /^#newdefine/ )
	{
	    chop;
	    ( $dummy, $key, $value ) = split ( /[ \t]+/, $_, 3 );

	    &Define ( "$key", $value ) if &GetValue($key) eq "";
	}
	elsif ( /^#timestamp[ \t]/ )
	{
	    chop;
	    ( $dummy, $stamp ) = split ( /[ \t]/, $_, 2 );

	    &FormatTimestamp;

	    &Define ( "TIMESTAMP", $stamp );
	}
	elsif ( ! /^#/ )
	{
	    &Substitute();
	    print OUTFILE $_;
	}
    }
    close $INFILE; --$filenum;

    $basePath = $saveBase;
    $tab = $saveTab;
}

# ---------------------------------------------------------------------

sub Error
{
    local ( $str ) = @_;

    print "   *** ERROR in line $.: $str\n";
}

# ---------------------------------------------------------------------

sub SetLinks
{
    &Define("TITLE", $ptitle[$ii]);

#   Go up one level.

    my $up_file;

    $i = $ii - 1;
    while ($i >= 0 && $plevel[$i] >= $plevel[$ii])
    {
	--$i;
    }

    if ($i >= 0 && $plevel[$i] < $plevel[$ii])
    {
	&DefineFile("LINK_UP", $pfile[$i]);
	&Define("TITLE_UP", $ptitle[$i]);
        $up_file = $pfile[$i];
    }
    else
    {
	&Undefine("LINK_UP");
	&Undefine("TITLE_UP");
    }

#   Next and previous documents.

    $i = $ii - 1;

    if ($i >= 0 && $pfile[$i] && $pfile[$i] ne $up_file)
    {
	&DefineFile("LINK_PREV", $pfile[$i]);
	&Define("TITLE_PREV", $ptitle[$i]);
    }
    else
    {
	&Undefine("LINK_PREV");
	&Undefine("TITLE_PREV");
    }

    $i = $ii + 1;

    if ($pfile[$i])
    {
	&DefineFile("LINK_NEXT", $pfile[$i]);
	&Define("TITLE_NEXT", $ptitle[$i]);
    }
    else
    {
	&Undefine("LINK_NEXT");
	&Undefine("TITLE_NEXT");
    }
}

# ---------------------------------------------------------------------

sub Define
{
    local ( $k, $v ) = @_;

    $v = "(((BLANK)))" if $v eq "";
    $defines{"$k"} = $v;

    @includePath = split(/:/,$v) if $k eq "INCLUDE_PATH";
    $outputDir = $v if $k eq "OUTPUT_DIR";
    $delim1 = $v if $k eq "OPEN_DELIMITER";
    $delim2 = $v if $k eq "CLOSE_DELIMITER";
}

# ---------------------------------------------------------------------

sub DefineFile
{
    local($f,$v) = @_;
    $v =~ s/\.$extSrc$/\.$extTgt/;
    &Define($f,$v);
}

# ---------------------------------------------------------------------

sub GetValue
{
    local($k) = @_;
    return $defines{"$k"};
}

# ---------------------------------------------------------------------

sub Undefine
{
    local($k) = @_;
    delete $defines{"$k"};
}

# ---------------------------------------------------------------------

sub Substitute
{
    $more = 1;

    while ( $more )
    {
	$p1 = index($_, $delim1);
	$p2 = index($_, $delim2);

	if ( $p1 >= 0 && $p2 >= 0 && $p2 > $p1 )
	{
	    $k = substr($_, $p1+2, $p2-$p1-2);
	    $v = $defines{$k};
	    $v = "" if $v eq "(((BLANK)))";
	    substr($_, $p1, $p2-$p1+2) = $v;
	}
	else
	{
	    $more = 0;
	}
    }

    if ( $entities )
    {
	s/&/&amp;/g;
	s/</&lt;/g;
	s/>/&gt;/g;
    }
}

# ---------------------------------------------------------------------

sub ResolveIncludeFile
{
    local($f) = $_[0];

    $file = $basePath . "/" . $f;
    $file = $f if $basePath eq "";

    return $f if -r $file;

    foreach $z (@includePath)
    {
	$file = $z . "/" . $f;
	return $file if -r $file;
    }
    &Error("No include file $f");
    return "";
}

# ---------------------------------------------------------------------

sub ResolvePath
{
    local ( $f ) = @_;
    local ( $full ) = $basePath . "/" . $f;

    $full = $f if $basePath eq "" || $f =~ /^\//;

#   Get rid of "./" sequences and convert "../" sequences.

    $prev = "";

    while ( $full =~ m|/\.| && $full != $prev )
    {
	$prev = $full;
	$full =~ s|//|/|g;
	$full =~ s|/\./|/|g;
	$full =~ s|[^./]+/\.\./||g;
    }

#   Now we just need the directory part and the file part.

    $i = rindex ( $full, "/" );

    if ( $i == -1 )
    {
	$basePath = "";
    }
    else
    {
	$basePath = substr ( $full, 0, $i );
    }
    return $full;
}

    
# ---------------------------------------------------------------------

# Get the current time and split it into the following printable values:
#
#   $sec	Seconds, 00 - 59
#   $min	Minutes, 00 - 59
#   $hour	Hours, 00 - 23
#   $mday	Day of the month, 1 - 31
#   $mon	Month number, 1 - 12
#   $monthname	Full month name, January - December
#   $shortmon	First three letters of month name, Jan - Dec
#   $year	Full year (e.g. 1996)
#   $syear	Last two digits of year (e.g. 96)

sub TimeNow
{
    @M = ( "January", "February", "March", "April", "May", "June",
	"July", "August", "September", "October", "November", "December" );

    ($sec,$min,$hour,$mday,$mon,$syear,$wday,$yday,$isdst) = localtime;

    $sec  = sprintf ( "%02d", $sec  );
    $min  = sprintf ( "%02d", $min  );
    $hour = sprintf ( "%02d", $hour );
    $year = $syear + 1900;

    $monthname = $M[$mon];
    $shortmon  = substr ( $monthname, 0, 3 );

    $mon++;			# Because it starts from 0
}

# ---------------------------------------------------------------------

# Create a printable time/date string based on a given format string.
#
# The format string is passed in the variable $stamp, and the following
# substitutions are made in it:
#
#   $ss		-> seconds
#   $mm		-> minutes
#   $hh		-> hour
#   $dd		-> day of the month
#   $MM		-> month number (1 - 12)
#   $Mmm	-> short month name (Jan - Dec)
#   $Month	-> full month name
#   $yyyy	-> full year (e.g. 1996)
#   $yy		-> short year (e.g. 96)
#
# Make sure you call "TimeNow" first to initialise the current time.

sub FormatTimestamp
{
    $stamp =~ s/\$ss/$sec/g;
    $stamp =~ s/\$mm/$min/g;
    $stamp =~ s/\$hh/$hour/g;
    $stamp =~ s/\$dd/$mday/g;
    $stamp =~ s/\$MM/$mon/g;
    $stamp =~ s/\$Month/$monthname/g;
    $stamp =~ s/\$Mmm/$shortmon/g;
    $stamp =~ s/\$yyyy/$year/g;
    $stamp =~ s/\$yy/$syear/g;
}

# ---------------------------------------------------------------------

sub AllSourceFiles
{
  my @files = ();
  my @dirs = (".");

  my $dir = pop @dirs;

  while ( $dir )
  {
    opendir(DIR, $dir) || die("Can't open $dir");

    my $file;
    foreach $file (grep !/^\./, readdir DIR)
    {
      if ( $file =~ /\.gtm$/ )
      {
	push @files, "$dir/$file";
      }
      elsif ( -d "$dir/$file" )
      {
	push @dirs, "$dir/$file";
      }
    }
    $dir = pop @dirs;
  }
  return @files;
}
